package DTDDoc;

import java.util.Hashtable;
import com.wutka.dtd.DTDComment;

/** <p>This class contains several tools used for parsing javadoc-like
 *  comments  (tags) appearing in XML comments. </p>
 *
 *  <p>Such a comment
 *  is made of text and "tags". A tag is a \@ immediately followed by
 *  a word and preceeded by at least a space.
 *  This function will return a hashtable that associate
 *  tags to their text. The text associated to a tag is the one
 *  that follows it and that precedes the next tag or the end
 *  of the comment. The text that appears before the first
 *  tag will automatically associated to the "comment" tag.</p>
 *  <p>Right now, each tag can appear once and only once !</p>
 *  <p>It is guaranteed that there are no spaces at the beginning or
 *  at the end of the tag's value.</p>
 *
 *  @author Stefan Champailler */
 
public class CommentParser {

    /** The character preceding each of the tag recognized in our comment
     *  system. */
     
    public static final char BEGIN_TAG = '@';
    
    /** The tag starting a comment. This one is put in front on the
     *  XML comments by default. */
     
    public static final String COMMENT_TAG = BEGIN_TAG+"comment";

    /** This function parses a javaDoc-like comment. The result is a
	 *  mapping between each tag and its value.
     *
     *  @param comment the comment
     *  @return a hashtable associating the comment tags to their value */

    private void parseComment( String comment) {

        //Hashtable r = new Hashtable();
        int start = 0;
        comment = COMMENT_TAG+" "+comment;

		//System.out.println("comment = "+comment);
        while( start < comment.length()) {

            int tagNameStart = comment.indexOf( BEGIN_TAG, start);

            // To work with "substring", the index of the last char
            // of the tagname is one char further than truth.
            int tagNameEnd = comment.indexOf( ' ', tagNameStart);

            if( tagNameStart >= start && tagNameEnd > tagNameStart) {
                int tagValueStart = tagNameEnd + 1;
                int tagValueEnd = comment.indexOf( BEGIN_TAG, tagValueStart);

                if( tagValueEnd == -1)
                    tagValueEnd = comment.length();

                String tagName = comment.substring( tagNameStart, tagNameEnd);

				/*if( "@attr".equalsIgnoreCase(tagName))
                	System.out.println("tagname = "+tagName);*/

                if( tagValueStart > tagNameEnd && tagValueEnd > tagValueStart) {
					
					// There's a value
					
                    String tagValue = comment.substring( tagValueStart, tagValueEnd).trim();
                
                	/*if( "@attr".equalsIgnoreCase(tagName))
						System.out.println("tagvalue = "+tagValue);*/
					    
                    if( tagValue == null) {
                    	
                    }
                    else if( tagValue.length() > 0 ) {
                    	//r.put( tagName, tagValue);
                    	putTagValue( tagName, tagValue);
					}

                    start = tagNameEnd + 1;
                } else {
					//start = comment.length();
					start = tagNameEnd + 1;
                }
                    
            } else
                start = comment.length();
        }

		//System.out.println("Done parsing ");
        //return r;
    }

	Hashtable values;
	
	String getUniqueTagValue( String tagName) {
		
		Object v = values.get( tagName);
			
		if( v == null)
			return null;
		else if( v instanceof String)
			return (String) v;
		else if( v instanceof Hashtable) {
			return null;
		} else {
			return null;
		}
	}

	Hashtable getMultipleTagValue( String tagName) {
		Object v = values.get( tagName);
			
		if( v == null)
			return null;
		else if( v instanceof Hashtable)
			return (Hashtable) v;
		else if( v instanceof String) {
			return null;
		} else {
			return null;
		}
	}
	
	void putTagValue( String name, String content) {

		if( name == null)
			throw new IllegalArgumentException("putTagValue(): null tag name !");
		
		if( "@attr".equalsIgnoreCase( name)) {

			if( content == null)
				throw new IllegalArgumentException("putTagValue(): content null for "+name+" !");

			String key = null;
			String value = null;
			int firstSpaceIndex = content.indexOf(' ');
		
			if( firstSpaceIndex != -1 && firstSpaceIndex+1 < content.length()) {
				key = content.substring( 0, firstSpaceIndex);
				value = content.substring( firstSpaceIndex+1, content.length());
			} else {
				return;
			}
			
			Hashtable current = (Hashtable) values.get( name);
			if( current == null) {
				current = new Hashtable();
				values.put( name, current);
			}
			current.put( key, value);
		} else {
			//System.out.println("putTagValue(): name="+name+" value="+content);
			values.put( name, content);
		}
	}

	CommentParser( DTDComment comment) {
		values = new Hashtable();
		parseComment( comment.getText());		
	}
}
