package com.japisoft.editix.editor.xsd.view2.nodeview;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;

import org.w3c.dom.Element;

import com.japisoft.editix.editor.xsd.view.View;
import com.japisoft.editix.editor.xsd.view2.node.XSDNode;
import com.japisoft.framework.preferences.Preferences;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class AbstractXSDNodeView implements XSDNodeView {

	protected XSDNode node;
	protected BufferedImage buffer = null;
	protected int hpadding = 10;
	protected int vpadding = 5;

	public final static Color DISABLED_COLOR = new Color( 200,200,200 );
	public final static Color BORDER_COLOR = new Color( 150, 150, 150 );
	
	public AbstractXSDNodeView( XSDNode node ) {
		this.node = node;
	}

	@Override
	public int getFullHeight( View designer ) {
		int height = getBuffer( designer ).getHeight();
		if ( node.getChildCount() > 0 ) {
			int childHeight = 0;
			for ( int i = 0; i < node.getChildCount(); i++ ) {
				if ( i > 0 ) {
					childHeight += 10;
				}
				childHeight += node.getChildAt( i ).getView().getFullHeight( designer );
			}			
			height = Math.max( height, childHeight );
		}
		return height;
	}

	@Override
	public int getHeight(View designer) {
		return getBuffer( designer ).getHeight();
	}
	
	@Override
	public int getWidth( View designer ) {
		return getBuffer( designer ).getWidth();
	}
	
	@Override
	public void paint( Graphics2D gc ) {
		if ( buffer != null )
			gc.drawImage( buffer, 0, 0, null );
	}

	@Override
	public void invalidateBuffer() {
		buffer = null;
	}

	private BufferedImage getBuffer( View designer ) {
		if ( buffer == null ) {
			createBuffer( designer );
		}
		return buffer;
	}
	
	protected BufferedImage createBufferedImage( int width, int height ) {
		return new BufferedImage(
			width + ( 2 * hpadding ),
			height + ( 2 * vpadding ),
			BufferedImage.TRANSLUCENT
		);
	}

	protected boolean isOptional() {
		Element e = node.getDOM();
		if ( e.hasAttribute( "minOccurs" ) ) {
			if ( "0".equals( e.getAttribute( "minOccurs" ) ) ) {
				return true;
			}	
		}
		return false;
	}

	// For occurences
	public void setBorder( Graphics2D gc ) {
		if ( isOptional() ) {
			// Dotted
			gc.setStroke(
				new BasicStroke(
					1, 
					BasicStroke.CAP_BUTT, 
					BasicStroke.JOIN_MITER, 
					3, 
					new float[]{3}, 
					0
				) 
			);
		}
	}
	
	protected Color getEnabledColor() {
		if ( node.isEnabled() ) {
			return ( Color.WHITE );
		} else {
			return ( DISABLED_COLOR );
		}		
	}
	
	public Color getSelectionColor() {
		return Preferences.getPreference( 
			"xsdEditor", 
			"selectionBackground", 
			new Color( 134, 150, 232 ) 
		);
	}

	protected void createBuffer( View designer ) {
		Graphics2D gc = ( Graphics2D )designer.getView().getGraphics();
		int width = gc.getFontMetrics().stringWidth( node.toString() );
		int height = gc.getFontMetrics().getHeight();

		buffer = createBufferedImage(
			width,
			height 
		);

		Graphics2D bgc = ( Graphics2D )buffer.getGraphics();
		
		if ( !node.isEnabled() || node.isReference() ) {
			bgc.setColor( DISABLED_COLOR );
		} else
		if ( node.isSelected() ) {
			bgc.setColor( getSelectionColor() );
		} else {				
			bgc.setColor( getEnabledColor() );
		}
		
		bgc.fillRoundRect( 0, 0, buffer.getWidth(), buffer.getHeight(), 5, 5 );
		
		bgc.setColor( Color.BLACK );
		
		bgc.setFont( gc.getFont() );
		bgc.drawString( node.toString(), hpadding, buffer.getHeight() - vpadding - gc.getFontMetrics().getDescent() );

		setBorder( bgc );

		if ( !node.isEnabled() || node.isReference() )
			bgc.setColor( DISABLED_COLOR );
		
		bgc.setColor( BORDER_COLOR );
		bgc.drawRoundRect( 0, 0, buffer.getWidth() - 1, buffer.getHeight() - 1, 5, 5 );
	}

}
