package com.japisoft.framework.xml.grammar.xsd;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.japisoft.framework.xml.DOMToolkit;
import com.japisoft.framework.xml.grammar.GrammarElement;
import com.japisoft.framework.xml.grammar.GrammarNode;
import com.japisoft.framework.xml.grammar.GrammarText;
import com.japisoft.framework.xml.grammar.GrammarType;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class XSDGrammarType implements GrammarType {

	private static final String SCHEMA_NS = "http://www.w3.org/2001/XMLSchema";

	private String nameType = null;

	public XSDGrammarType( String type ) {
		this.nameType = type;
	}

	private Document doc = null;
	private XSDGrammar grammar = null;
	
	public XSDGrammarType( XSDGrammar grammar, Element definitionType ) {
		this.grammar = grammar;
		doc = definitionType.getOwnerDocument();
		try {
			processElement( definitionType );
		} finally {
			doc = null;
		}
	}

	public String infer() {
		if ( "boolean".equals( nameType ) ) {
			return "false";
		} else
		if ( "float".equals( nameType ) ) {
			return "1.0";
		} else
		if ( "double".equals( nameType ) ) {
			return "1.0";
		} else
		if ( "integer".equals( nameType ) ) {
			return "1";
		} else
		if ( "nonPositiveInteger".equals( nameType ) )
			return "-1";
		else
		if ( "negativeInteger".equals( nameType ) )
			return "-1";
		else
		if ( "long".equals( nameType ) ) {
			return "1";
		} else
		if ( "byte".equals( nameType ) ) {
			return "1";
		} else
		if ( "nonNegativeInteger".equals( nameType ) ) {
			return "1";
		} else
		if ( "unsignedLong".equals( nameType ) ) {
			return "1";
		} else
		if ( "unsignedInt".equals( nameType ) ) {
			return "1";
		} else
		if ( "unsignedShort".equals( nameType ) ) {
			return "1";
		} else
		if ( "unsignedByte".equals( nameType ) ) {
			return "1";
		} else
		if ( "positiveInteger".equals( nameType ) ) {
			return "1";
		}
		return "";
	}

	public XSDGrammar getGrammar() {
		return grammar;
	}

	private void processElement( Element definitionType ) {

		Element refType = 
			definitionType;

		if ( definitionType.hasAttribute( "ref" ) ) {
			definitionType = resolveRef( definitionType );
			if ( definitionType == null ) {
				// Can't find the reference
				definitionType = refType;
			}
		}

		nameType = 
			definitionType.getAttribute( "type" );

		if ( definitionType.hasAttribute( "base" ) )
			nameType = definitionType.getAttribute( "base" );

		Element explorableType = null;

		// Look at children in first
		NodeList nl = definitionType.getChildNodes();
		for  ( int i = 0; i < nl.getLength(); i++ ) {
			if ( !( nl.item( i ) instanceof Element ) )
				continue;

			Element e = ( Element )nl.item( i );
			if ( "simpleType".equals( e.getLocalName() ) ) { 
					explorableType = e;
					break;
			} else 
				if ( "complexType".equals( e.getLocalName() ) ) {
					explorableType = e;
					break;
				}
		}

		if ( explorableType == null && 
				!"".equals( nameType ) ) {
			// Search this type
			explorableType = 
				resolveType( nameType );
		}

		if ( explorableType != null ) {

			if ( "simpleType".equals( 
					explorableType.getLocalName() ) )
				processSimpleType( 
						explorableType );
			else
				processComplexType( 
						refType, 
						explorableType );

		} else {

			// If we can't resolve it then this is a default one
			nameType = extractLocalType(
					definitionType.getAttribute( "type" ), 
					definitionType );

		}

	}

	private void processComplexContent( Element complexContent ) {

		Element e = 
			DOMToolkit.getFirstElement( 
				complexContent, 
				"restriction" );

		if ( e == null ) {
			e = DOMToolkit.getFirstElement( 
				complexContent, 
				"extension" );
		}			

		if ( e != null ) {
			
			String type = e.getAttribute( "base" );

			Element resolveBase = resolveType( type );
			if ( resolveBase != null ) {

				processElement( resolveBase );

			}

			processComplexParts( complexContent );
		}
		
	}

	private String processSimpleContent( Element simpleContent ) {
		Element e = 
			DOMToolkit.getFirstElement( 
				simpleContent, 
				"restriction" );

		if ( e == null ) {
			e = DOMToolkit.getFirstElement( 
				simpleContent, 
				"extension" );
		}			

		if ( e != null ) {
			String type = e.getAttribute( "base" );
			Element resolveBase = resolveType( type );

			if ( resolveBase != null ) {

				processSimpleType( resolveBase );

			} else {

				String resType = 
					extractLocalType( type, e );

				if ( resType != null ) {

					GrammarNode node = 
						getFirstTextGrammarNode();

					if ( node != null ) {

						removeGrammarNode( node );
						
					}

					addGrammarNodeValue( 
							new XSDGrammarText(
									null, 
									new XSDGrammarType( 
											resType )  ) );

				}

			}

			processComplexParts( e );

			return type;
		}

		return null;
	}

	/** For complexType, simpleContent, complexContent */
	void processComplexParts( Element complexPart ) {

		NodeList nlext = complexPart.getChildNodes();

		if ( nlext.getLength() > 0 ) {
			for ( int i = 0; i < nlext.getLength(); i++ ) {
				Node n = nlext.item( i );
				if ( n instanceof Element ) {
					if ( SCHEMA_NS.equals( 
							n.getNamespaceURI() ) ) {
						if ( "sequence".equals( n.getLocalName() ) ||
								"choice".equals( n.getLocalName() ) ||
									"all".equals( n.getLocalName() ) ) {	
							addGrammarNodeValue(
									new XSDGrammarContainer( grammar, ( Element )n ) );
						} else
						if ( "element".equals( n.getLocalName() ) ) {
							addElementValue( ( Element )n );
						} else
						if ( "attribute".equals( n.getLocalName() ) ) {
							addAttributeValue( ( Element )n );
						}
					}
				}
			}
		}
	}

	void processSimpleType(
			Element simpleType ) {

		Element e = 
			DOMToolkit.getFirstElement( 
				simpleType, 
				"restriction" );

		if ( e == null ) {
			e = DOMToolkit.getFirstElement( 
				simpleType, 
				"extension" );
		}			

		if ( e != null ) {

			// Explore the parent type
			
			nameType = e.getAttribute( "base" );
			Element resolveBase = resolveType( nameType );
			if ( resolveBase != null ) {

				processSimpleContent( resolveBase );

			}

		} else {
			 
			e = DOMToolkit.getFirstElement( simpleType, "list" );
			if ( e != null ) {
				nameType = "list";
			} else {
				e = DOMToolkit.getFirstElement( simpleType, "union" );
				if ( e != null ) {
					nameType = "union";
				}				
			}

			
		}

		// Get all the enumeration

		NodeList nlEnum = simpleType.getElementsByTagNameNS( SCHEMA_NS, "enumeration" );

		for ( int i = 0; i < nlEnum.getLength(); i++ ) {
			addEnumValue( 
					( Element )nlEnum.item( i ) 
			);
		}

	}

	public boolean isComplex() {
		return "complex".equals( nameType );
	}

	void processComplexType( 
			Element refType, 
			Element complexType ) {
		
		nameType = "complex";		

		Element sc = 
			DOMToolkit.getFirstElement(
					complexType, 
					"simpleContent" );

		if ( sc != null ) {

			processSimpleContent( sc );

		} else {

			sc = 
				DOMToolkit.getFirstElement(
						complexType, 
						"complexContent" );

			if ( sc != null ) {

				processComplexContent( sc );

			} else {

				processComplexParts( complexType );

			}

		}

	}

/*	private Element resolveType( String type ) {
		List<GrammarElement> l = 
			XSDGrammar.getGlobalComponent( 
					doc, "simpleType" );
		for ( GrammarElement ge : l ) {
			Element e = ( ( XSDGrammarElement )ge ).getDOMElement();
			if ( e.getAttribute( "name" ).equals( type ) )
				return e;
		}
		l = 
			XSDGrammar.getGlobalComponent( 
					doc, "complexType" );
		for ( GrammarElement ge : l ) {
			Element e = ( ( XSDGrammarElement )ge ).getDOMElement();
			if ( e.getAttribute( "name" ).equals( type ) )
				return e;
		}
		return null;
	} */

	private Element resolveType( String definitionType ) {
		NodeList nl = 
			doc.getDocumentElement().getChildNodes();
		for ( int i = 0; i < nl.getLength(); i++ ) {
			if ( nl.item( i ) instanceof Element ) {
				Element e = ( Element )nl.item( i );
				if ( "simpleType".equals( e.getLocalName() ) ||
						"complexType".equals( e.getLocalName() ) ) {
					if ( definitionType.equals( e.getAttribute( "name" ) ) ) {
						return e;
					}
				}
			}
		}
		return null;
	}

	private Element resolveRef( Element definitionType ) {
		String ref = definitionType.getAttribute( "ref" );
		List<GrammarElement> l =  XSDGrammar.getGlobalComponent(
				grammar,
				doc, 
				"element" 
		);
		for ( GrammarElement ge : l ) {
			Element e = ( ( XSDGrammarElement )ge ).getDOMElement();
			if ( e.getAttribute( "name" ).equals( ref ) )
				return e;
		}
		// Not found ??
		return null;
	}

	public String getType() {
		return nameType;
	}

	private void addEnumValue( Element e ) {
		addGrammarNodeValue( 
				new XSDGrammarText(
						grammar,
						e.getAttribute( "value" ) ) );
	}

	private void addAttributeValue( Element e ) {
		addGrammarNodeValue(
			new XSDGrammarAttribute( grammar, e ) 
		);		
	}

	private void addElementValue( Element e ) {
		addGrammarNodeValue( 
			new XSDGrammarElement( grammar, e ) );		
	}

	private void addGrammarNodeValue( GrammarNode e ) {
		if ( children == null )
			children = new ArrayList<GrammarNode>();
		children.add( e );
	}
	
	private void removeGrammarNode( GrammarNode e ) {
		if ( children != null ) {
			children.remove( e );
		}
	}

	private GrammarText getFirstTextGrammarNode() {
		if ( children == null )
			return null;
		for ( GrammarNode node : children ) {
			if ( node instanceof GrammarText )
				return ( GrammarText )node;
		}
		return null;
	}

	private ArrayList<GrammarNode> children = null;

	public List<GrammarNode> getValues() {
		return children;
	}

	private String extractLocalType( String type, Element e ) {
		if ( type != null ) {
			String prefix = getRootPrefix( e );
			if ( prefix != null ) {
				if ( type.startsWith( 
						prefix + ":" ) ) {
					type = type.substring( 
							prefix.length() + 1 );
					return type;
				}
			}
		}
		return null;
	}

	protected String getRootPrefix( Node node ) {
		Document doc = node.getOwnerDocument();
		Element root = doc.getDocumentElement();
		return root.getPrefix();
	}	

}
