package com.japisoft.framework.xml.parser.sax;

import java.io.*;
import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import com.japisoft.framework.xml.parser.HandlerException;
import com.japisoft.framework.xml.parser.ParseException;
import com.japisoft.framework.xml.parser.FPParser;
import com.japisoft.framework.xml.parser.document.DocumentBuilder;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public final class SaxParser extends FPParser implements org.xml.sax.Parser,
		Locator {

	public SaxParser() {
		super();
		setEnabledNameSpace(false);
		bufferingMode(false);
	}

	/**
	 * Allow an application to request a locale for errors and warnings.
	 * 
	 * <p>
	 * SAX parsers are not required to provide localisation for errors and
	 * warnings; if they cannot support the requested locale, however, they must
	 * throw a SAX exception. Applications may not request a locale change in
	 * the middle of a parse.
	 * </p>
	 * 
	 * @param locale
	 *            A Java Locale object.
	 * @exception org.xml.sax.SAXException
	 *                Throws an exception (using the previous or default locale)
	 *                if the requested locale is not supported.
	 * @see org.xml.sax.SAXException
	 * @see org.xml.sax.SAXParseException
	 */
	public void setLocale(Locale locale) throws SAXException {
		throw new SAXException("Not supported");
	}

	/**
	 * Allow an application to register a custom entity resolver.
	 * 
	 * <p>
	 * If the application does not register an entity resolver, the SAX parser
	 * will resolve system identifiers and open connections to entities itself
	 * (this is the default behaviour implemented in HandlerBase).
	 * </p>
	 * 
	 * <p>
	 * Applications may register a new or different entity resolver in the
	 * middle of a parse, and the SAX parser must begin using the new resolver
	 * immediately.
	 * </p>
	 * 
	 * @param resolver
	 *            The object for resolving entities.
	 * @see EntityResolver
	 * @see HandlerBase
	 */
	public void setEntityResolver(EntityResolver resolver) {
		// No effect
	}

	/**
	 * Allow an application to register a DTD event handler.
	 * 
	 * <p>
	 * If the application does not register a DTD handler, all DTD events
	 * reported by the SAX parser will be silently ignored (this is the default
	 * behaviour implemented by HandlerBase).
	 * </p>
	 * 
	 * <p>
	 * Applications may register a new or different handler in the middle of a
	 * parse, and the SAX parser must begin using the new handler immediately.
	 * </p>
	 * 
	 * @param handler
	 *            The DTD handler.
	 * @see DTDHandler
	 * @see HandlerBase
	 */
	public void setDTDHandler(DTDHandler handler) {
		// No effect
	}

	/**
	 * Allow an application to register a document event handler.
	 * 
	 * <p>
	 * If the application does not register a document handler, all document
	 * events reported by the SAX parser will be silently ignored (this is the
	 * default behaviour implemented by HandlerBase).
	 * </p>
	 * 
	 * <p>
	 * Applications may register a new or different handler in the middle of a
	 * parse, and the SAX parser must begin using the new handler immediately.
	 * </p>
	 * 
	 * @param handler
	 *            The document handler.
	 * @see DocumentHandler
	 * @see HandlerBase
	 */
	public void setDocumentHandler(DocumentHandler handler) {
		this.handler = handler;
		handler.setDocumentLocator(this);
	}

	private DocumentHandler handler;

	/**
	 * Allow an application to register an error event handler.
	 * 
	 * <p>
	 * If the application does not register an error event handler, all error
	 * events reported by the SAX parser will be silently ignored, except for
	 * fatalError, which will throw a SAXException (this is the default
	 * behaviour implemented by HandlerBase).
	 * </p>
	 * 
	 * <p>
	 * Applications may register a new or different handler in the middle of a
	 * parse, and the SAX parser must begin using the new handler immediately.
	 * </p>
	 * 
	 * @param handler
	 *            The error handler.
	 * @see ErrorHandler
	 * @see SAXException
	 * @see HandlerBase
	 */
	public void setErrorHandler(ErrorHandler handler) {
		this.errorHandler = handler;
	}

	private ErrorHandler errorHandler;

	/**
	 * Parse an XML document.
	 * 
	 * <p>
	 * The application can use this method to instruct the SAX parser to begin
	 * parsing an XML document from any valid input source (a character stream,
	 * a byte stream, or a URI).
	 * </p>
	 * 
	 * <p>
	 * Applications may not invoke this method while a parse is in progress
	 * (they should create a new Parser instead for each additional XML
	 * document). Once a parse is complete, an application may reuse the same
	 * Parser object, possibly with a different input source.
	 * </p>
	 * 
	 * @param source
	 *            The input source for the top-level of the XML document.
	 * @exception org.xml.sax.SAXException
	 *                Any SAX exception, possibly wrapping another exception.
	 * @exception java.io.IOException
	 *                An IO exception from the parser, possibly from a byte
	 *                stream or character stream supplied by the application.
	 * @see org.xml.sax.InputSource
	 * @see #parse(java.lang.String)
	 * @see #setEntityResolver
	 * @see #setDTDHandler
	 * @see #setDocumentHandler
	 * @see #setErrorHandler
	 */
	public void parse(InputSource source) throws SAXException, IOException {
		systemId = source.getSystemId();
		try {
			start = false;
			parse( source.getCharacterStream());
			if (handler != null) {
				handler.endDocument();
			}
		} catch (ParseException e) {
			if (errorHandler != null) {
				errorHandler.error(new SAXParseException(e.getMessage(), null,
						null, e.getLine(), e.getCol()));
			}
			throw new SAXException(e.getMessage(), e);
		}
	}
	
	/**
	 * Parse an XML document from a system identifier (URI).
	 * 
	 * <p>
	 * This method is a shortcut for the common case of reading a document from
	 * a system identifier. It is the exact equivalent of the following:
	 * </p>
	 * 
	 * <pre>
	 * parse(new InputSource(systemId));
	 * </pre>
	 * 
	 * <p>
	 * If the system identifier is a URL, it must be fully resolved by the
	 * application before it is passed to the parser.
	 * </p>
	 * 
	 * @param systemId
	 *            The system identifier (URI).
	 * @exception org.xml.sax.SAXException
	 *                Any SAX exception, possibly wrapping another exception.
	 * @exception java.io.IOException
	 *                An IO exception from the parser, possibly from a byte
	 *                stream or character stream supplied by the application.
	 * @see #parse(org.xml.sax.InputSource)
	 */
	public void parse(String systemId) throws IOException, SAXException {
		parse(new InputSource(this.systemId = systemId));
	}

	/////////////////////////////////////////////////////////////////////////////

	private boolean start = false;

	private String currentTag;

	private String currentAttribute;

	private AttributeListImpl currentAttributes;

	protected void fireItemFound(DocumentBuilder db, int state, String item) throws ParseException {
		if (handler == null)
			super.fireItemFound(db, state, item);
		else {
			try {
				if (!start) {
					handler.startDocument();
					start = true;
				}

				String tmpCurrentTag = null;

				if (currentTag != null) {
					if ((state != 5) && (state != 8)) {

						if (currentAttributes == null)
							currentAttributes = new AttributeListImpl();

						// Open tag
						handler.startElement(currentTag, currentAttributes);
						tmpCurrentTag = currentTag;
						currentTag = null;
						currentAttributes = null;
					}
				}

				switch (state) {
				case 2: // TEXT
					handler.characters(item.toCharArray(), 0, item.length());
					break;
				case 3: {
					if (item.startsWith("/")) {
						handler.endElement(item.substring(1));
					} else {
						if (!item.startsWith("?")) {
							if (!"".equals(item)) {
								currentTag = item;
								currentAttributes = null;
							}
						}
					}
					break;
				}
				case 5: // Attribute name
					currentAttribute = item;
					break;
				case 8: // Attribute value
					if (currentAttributes == null) {
						currentAttributes = new AttributeListImpl();
					}
					currentAttributes
							.addAttribute(currentAttribute, null, item);
					currentAttribute = null;
					break;
				case 10:
					if ("/".equals(item)) {
						if (tmpCurrentTag == null)
							handler.endElement(currentTag);
						else
							handler.endElement(tmpCurrentTag);
						tmpCurrentTag = null;
						currentTag = null;
						currentAttributes = null;
					}
					break;
				case 13:
					// IGNORE COMMENT ?
					break;
				case 15: // CDATA
					item = item
							.substring("[CDATA[".length() - 1, item.length());
					handler.characters(item.toCharArray(), 0, item.length());
					break;
				}
			} catch (SAXException e) {
				throw new ParseException(e.getMessage());
			} catch (Throwable th) {
				throw new HandlerException(th);
			}
		}
	}

	//////////////////// LOCATOR PART ///////////////////

	private String systemId = null;

	public int getColumnNumber() {
		return col;
	}

	public int getLineNumber() {
		return line;
	}

	public String getPublicId() {
		return null;
	}

	public String getSystemId() {
		return systemId;
	}

	/////////////////////////////////////////////////////

	public static void main(String[] args) throws Exception {
		SaxParser sp = new SaxParser();
		sp.parse(new FileInputStream("/tmp/test.xml"));
		System.out.println("Ok");
	}
	
}

// SaxParser ends here
