package com.japisoft.xflows.task;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.swing.JDialog;
import javax.swing.JTextField;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class FilesTaskRunner implements TaskRunner {
	
	protected TaskRunner monoTask;
	
	public static final String SOURCEPATH = "sourcepath";
	public static final String SOURCEFILTER = "sourcefilter";
	public static final String TARGETPATH = "targetpath";
	public static final String TARGETNAME = "targetname";

	private boolean targetRequired = false;

	public FilesTaskRunner( TaskRunner monoTask, boolean targetRequired ) {
		this.monoTask = monoTask;
		this.targetRequired = targetRequired;
	}

	protected boolean fileTarget = false;
	protected boolean defaultProcessingLog = false;
	
	public boolean run( TaskContext context ) {

		// Check source path
		String sourcePath = context.getParam( SOURCEPATH );
		if ( sourcePath == null ) {
			context.addError( "No source path " );
			return ERROR;
		}
		File fsource = new File( sourcePath );
		if ( !fsource.exists() ) {
			context.addError( "Cannot find the source path : " + sourcePath );
			return ERROR;
		}

		String sourceFilter = null;
		boolean sourceDirectory = false;
		
		if ( fsource.isDirectory() ) {
			sourceDirectory = true;
			sourceFilter = context.getParam( SOURCEFILTER );
			if ( sourceFilter == null ) {
				sourceFilter = "(.*)\\.(.*)";
				context.addWarning( "No sourceFilter found : Use the default one " + sourceFilter );
			}
		}

		// Check target path
		String targetPath = context.getParam( TARGETPATH );
		String targetName = null;
		File ftarget = null;
		
		if ( targetPath != null ) {
			ftarget = new File( targetPath );

			if ( ftarget.isDirectory() ) {
				
				if  ( !sourceDirectory ) {
					context.addError( "Invalid target, a file name is required" );
					return ERROR;
				}

				// It requires a targetname
				targetName = context.getParam( TARGETNAME );
				if ( targetName == null ) {
					context.addError( "Target name not found " );
					return ERROR;
				}
			} else {
				
				if ( !fileTarget && context.hasParam( TARGETNAME ) ) {

					context.addWarning( "Target name defined but no directory found for the target" );

				}
				
			}
				

		} else {
			if ( targetRequired ) {
				context.addError( "No target path" );
				return ERROR;
			}
		}

		if ( !sourceDirectory ) {
			context.setCurrentSourceFile( fsource );
			context.setCurrentTargetFile( ftarget );
			if ( defaultProcessingLog )
				context.addInfo( "Processing " + context.getCurrentSourceFile() );
			return monoTask.run( context );
		} else {

			String[] content = fsource.list();
			boolean errorFound = false;
			
			if ( content != null ) {
				try {
					boolean processOnce = false;
					/*StringBuffer sb = new StringBuffer();
					for ( int i = 0; i < sourceFilter.length(); i++ ) {
						if ( sourceFilter.charAt( i ) == '\\' )
							sb.append( "\\\\" );
						else
							sb.append( sourceFilter.charAt( i ) );
					} */					
					Pattern p = Pattern.compile( sourceFilter );
					for ( int i = 0; i < content.length; i++ ) {
						if ( context.isInterrupted() )
							break;

						String name = content[ i ];						
						Matcher m = p.matcher( name );

						if ( m.matches() ) {
							processOnce = true;
							String finalTarget = targetName;
							if  ( targetName != null ) {								
								for ( int j = 0; j < m.groupCount() ;j++ ) {
									String s = m.group( j + 1 );
									finalTarget = finalTarget.replaceAll( "\\$" + ( j + 1 ), s );
								}
							}
							context.setCurrentSourceFile( new File( fsource, name ) );
							if ( targetName != null && ftarget != null ) {
								context.setCurrentTargetFile( new File( ftarget, finalTarget ) );
							} else
							if ( ftarget != null ) {
								context.setCurrentTargetFile( ftarget );
							}

							if ( defaultProcessingLog )
								context.addInfo( "Processing " + context.getCurrentSourceFile() );
							
							if ( monoTask.run( context ) == ERROR )
								errorFound = true;
						}						
					}

					if ( !processOnce )
						context.addWarning( "No source file found" );

				} catch( PatternSyntaxException exc ) {
					context.addError( "Wrong source filter " + sourceFilter + " : " + exc.getMessage( ) );
				}
				
				if ( errorFound )
					return ERROR;
				
			} else
				context.addWarning( "No content for directory " + fsource );
		}		
		return OK;
	}

	public static void main( String[] args ) {
		
		JDialog dialog = new JDialog();
		dialog.setModal( true );
		JTextField tf = new JTextField();
		dialog.getContentPane().add( tf );
		dialog.setSize( 400, 30 );
		dialog.setVisible( true );

		Pattern p = Pattern.compile( tf.getText() );
		Matcher m = p.matcher( "test.toto" );
		System.out.println( "Match = " + m.matches() );
	}
	
}
